@@ -21,6 +21,7 @@ from swift.common.utils import public, Timestamp, \
     override_bytes_from_content_type
 from swift.common.constraints import check_metadata
 from swift.common import constraints
+from swift.common.middleware.versioned_writes import DELETE_MARKER_CONTENT_TYPE
 from swift.common.swob import Response, HTTPBadRequest, HTTPNotFound, \
     HTTPNoContent, HTTPConflict, HTTPPreconditionFailed, HTTPForbidden, \
     HTTPCreated
@@ -29,12 +30,13 @@ from swift.common.request_helpers import is_sys_or_user_meta, get_param
 from swift.proxy.controllers.container import ContainerController \
         as SwiftContainerController
 from swift.proxy.controllers.base import clear_info_cache, \
-    delay_denial, cors_validation, _set_info_cache
+    delay_denial, cors_validation, set_info_cache
 
 from oio.common import exceptions
 
-from oioswift.common.storage_policy import POLICIES
-from oioswift.utils import get_listing_content_type
+from oioswift.utils import get_listing_content_type, \
+    handle_oio_no_such_container, handle_oio_timeout, \
+    handle_service_busy
 
 
 class ContainerController(SwiftContainerController):
@@ -43,7 +45,11 @@ class ContainerController(SwiftContainerController):
                             'x-container-sync-key', 'x-container-sync-to',
                             'x-versions-location']
 
+    @handle_oio_no_such_container
+    @handle_oio_timeout
+    @handle_service_busy
     def GETorHEAD(self, req):
+        """Handler for HTTP GET/HEAD requests."""
         if self.account_info(self.account_name, req) is None:
             if 'swift.authorize' in req.environ:
                 aresp = req.environ['swift.authorize'](req)
@@ -55,8 +61,8 @@ class ContainerController(SwiftContainerController):
             resp = self.get_container_list_resp(req)
         else:
             resp = self.get_container_head_resp(req)
-        _set_info_cache(self.app, req.environ, self.account_name,
-                        self.container_name, resp)
+        set_info_cache(self.app, req.environ, self.account_name,
+                       self.container_name, resp)
         resp = self.convert_policy(resp)
         if 'swift.authorize' in req.environ:
             req.acl = resp.headers.get('x-container-read')
@@ -72,7 +78,7 @@ class ContainerController(SwiftContainerController):
     def convert_policy(self, resp):
         if 'X-Backend-Storage-Policy-Index' in resp.headers and \
                 is_success(resp.status_int):
-                    policy = POLICIES.get_by_index(
+                    policy = self.app.POLICIES.get_by_index(
                         resp.headers['X-Backend-Storage-Policy-Index'])
                     if policy:
                         resp.headers['X-Storage-Policy'] = policy.name
@@ -85,11 +91,7 @@ class ContainerController(SwiftContainerController):
 
     def get_metadata_resp_headers(self, meta):
         headers = {}
-        try:
-            system = meta['system']
-        except KeyError:
-            # compatibility with oio-sds < 3.2
-            system = meta['properties']
+        system = meta.get('system') or {}
         # sys.m2.ctime is microseconds
         ctime = float(system.get('sys.m2.ctime', 0)) / 1000000.0
         headers.update({
@@ -106,8 +108,6 @@ class ContainerController(SwiftContainerController):
         return headers
 
     def get_container_list_resp(self, req):
-        storage = self.app.storage
-
         path = get_param(req, 'path')
         prefix = get_param(req, 'prefix')
         delimiter = get_param(req, 'delimiter')
@@ -132,23 +132,24 @@ class ContainerController(SwiftContainerController):
             if path:
                 prefix = path.rstrip('/') + '/'
             delimiter = '/'
+        opts = req.environ.get('oio.query', {})
+        oio_headers = {'X-oio-req-id': self.trans_id}
+        result = self.app.storage.object_list(
+            self.account_name, self.container_name, prefix=prefix,
+            limit=limit, delimiter=delimiter, marker=marker,
+            end_marker=end_marker, properties=True,
+            versions=opts.get('versions', False),
+            deleted=opts.get('deleted', False),
+            headers=oio_headers)
 
-        try:
-            result = storage.object_list(
-                self.account_name, self.container_name, prefix=prefix,
-                limit=limit, delimiter=delimiter, marker=marker,
-                end_marker=end_marker, properties=True)
-
-            resp_headers = self.get_metadata_resp_headers(result)
-            resp = self.create_listing(
-                req, out_content_type, resp_headers, result,
-                self.container_name)
-        except exceptions.NoSuchContainer:
-            return HTTPNotFound(request=req)
+        resp_headers = self.get_metadata_resp_headers(result)
+        resp = self.create_listing(
+            req, out_content_type, resp_headers, result,
+            self.container_name, **opts)
         return resp
 
     def create_listing(self, req, out_content_type, resp_headers,
-                       result, container):
+                       result, container, **kwargs):
         container_list = result['objects']
         for p in result.get('prefixes', []):
             record = {'name': p,
@@ -156,13 +157,14 @@ class ContainerController(SwiftContainerController):
             container_list.append(record)
         ret = Response(request=req, headers=resp_headers,
                        content_type=out_content_type, charset='utf-8')
+        versions = kwargs.get('versions', False)
         if out_content_type == 'application/json':
-            ret.body = json.dumps([self.update_data_record(r)
-                                   for r in container_list])
+            ret.body = json.dumps(
+                [self.update_data_record(r, versions) for r in container_list])
         elif out_content_type.endswith('/xml'):
             doc = Element('container', name=container.decode('utf-8'))
             for obj in container_list:
-                record = self.update_data_record(obj)
+                record = self.update_data_record(obj, versions)
                 if 'subdir' in record:
                     name = record['subdir'].decode('utf-8')
                     sub = SubElement(doc, 'subdir', name=name)
@@ -186,7 +188,7 @@ class ContainerController(SwiftContainerController):
 
         return ret
 
-    def update_data_record(self, record):
+    def update_data_record(self, record, versions=False):
         if 'subdir' in record:
             return {'subdir': record['name']}
 
@@ -196,6 +198,10 @@ class ContainerController(SwiftContainerController):
                     'last_modified': Timestamp(record['ctime']).isoformat,
                     'content_type': record.get(
                         'mime_type', 'application/octet-stream')}
+        if record.get('deleted', False):
+            response['content_type'] = DELETE_MARKER_CONTENT_TYPE
+        if versions:
+            response['version'] = record.get('version', 'null')
         override_bytes_from_content_type(response)
         return response
 
@@ -217,31 +223,35 @@ class ContainerController(SwiftContainerController):
         headers = dict()
         out_content_type = get_listing_content_type(req)
         headers['Content-Type'] = out_content_type
-        storage = self.app.storage
-        try:
-            meta = storage.container_get_properties(self.account_name,
-                                                    self.container_name)
-            headers.update(self.get_metadata_resp_headers(meta))
-            resp = HTTPNoContent(request=req, headers=headers, charset='utf-8')
-        except exceptions.NoSuchContainer:
-            resp = HTTPNotFound(request=req, headers=headers)
+        oio_headers = {'X-oio-req-id': self.trans_id}
+        meta = self.app.storage.container_get_properties(
+            self.account_name, self.container_name, headers=oio_headers)
+        headers.update(self.get_metadata_resp_headers(meta))
+        return HTTPNoContent(request=req, headers=headers, charset='utf-8')
 
-        return resp
-
-    def load_container_metadata(self, headers):
-        metadata = {}
-        metadata.update(
-            (k, v)
-            for k, v in headers.iteritems()
+    def properties_from_headers(self, headers):
+        metadata = {
+            k: v
+            for k, v in headers.items()
             if k.lower() in self.pass_through_headers or
-            is_sys_or_user_meta('container', k))
-        return metadata
+            is_sys_or_user_meta('container', k)
+        }
+
+        system = dict()
+        # This header enables versioning
+        ver_loc = headers.get('X-Container-Sysmeta-Versions-Location')
+        if ver_loc is not None:
+            # When suspending versioning, header has empty string value
+            ver_val = "-1" if ver_loc else "1"
+            system['sys.m2.policy.version'] = ver_val
+
+        return metadata, system
 
     def _convert_policy(self, req):
         policy_name = req.headers.get('X-Storage-Policy')
         if not policy_name:
             return
-        policy = POLICIES.get_by_name(policy_name)
+        policy = self.app.POLICIES.get_by_name(policy_name)
         if not policy:
             msg = "Invalid X-Storage-Policy '%s'" % policy_name
             raise HTTPBadRequest(
@@ -249,11 +259,13 @@ class ContainerController(SwiftContainerController):
         return policy
 
     def get_container_create_resp(self, req, headers):
-        metadata = self.load_container_metadata(headers)
+        properties, system = self.properties_from_headers(headers)
         # TODO container update metadata
-        storage = self.app.storage
-        created = storage.container_create(
-            self.account_name, self.container_name, properties=metadata)
+        oio_headers = {'X-oio-req-id': self.trans_id}
+        created = self.app.storage.container_create(
+            self.account_name, self.container_name,
+            properties=properties, system=system,
+            headers=oio_headers)
         if created:
             return HTTPCreated(request=req)
         else:
@@ -261,6 +273,8 @@ class ContainerController(SwiftContainerController):
 
     @public
     @cors_validation
+    @handle_oio_timeout
+    @handle_service_busy
     def PUT(self, req):
         """HTTP PUT request handler."""
         error_response = \
@@ -305,6 +319,8 @@ class ContainerController(SwiftContainerController):
 
     @public
     @cors_validation
+    @handle_oio_timeout
+    @handle_service_busy
     def POST(self, req):
         """HTTP POST request handler."""
         error_response = \
@@ -327,43 +343,45 @@ class ContainerController(SwiftContainerController):
         return resp
 
     def get_container_post_resp(self, req, headers):
-        storage = self.app.storage
-
-        metadata = self.load_container_metadata(headers)
-        if not metadata:
+        properties, system = self.properties_from_headers(headers)
+        if not properties:
             return self.PUT(req)
 
+        oio_headers = {'X-oio-req-id': self.trans_id}
         try:
-            storage.container_set_properties(
-                self.account_name, self.container_name, metadata)
+            self.app.storage.container_set_properties(
+                self.account_name, self.container_name,
+                properties=properties, system=system,
+                headers=oio_headers)
             resp = HTTPNoContent(request=req)
         except exceptions.NoSuchContainer:
             resp = self.PUT(req)
         return resp
 
-    def get_container_delete_resp(self, req, headers):
-        storage = self.app.storage
+    def get_container_delete_resp(self, req):
+        oio_headers = {'X-oio-req-id': self.trans_id}
         try:
-            storage.container_delete(self.account_name, self.container_name)
+            self.app.storage.container_delete(
+                self.account_name, self.container_name, headers=oio_headers)
         except exceptions.ContainerNotEmpty:
             return HTTPConflict(request=req)
-        except exceptions.NoSuchContainer:
-            return HTTPNotFound(request=req)
         resp = HTTPNoContent(request=req)
         return resp
 
     @public
     @cors_validation
+    @handle_oio_no_such_container
+    @handle_oio_timeout
+    @handle_service_busy
     def DELETE(self, req):
         """HTTP DELETE request handler."""
         account_partition, accounts, container_count = \
             self.account_info(self.account_name, req)
         if not accounts:
             return HTTPNotFound(request=req)
-        headers = self.generate_request_headers(req, transfer=True)
         clear_info_cache(self.app, req.environ,
                          self.account_name, self.container_name)
-        resp = self.get_container_delete_resp(req, headers)
+        resp = self.get_container_delete_resp(req)
         if resp.status_int == HTTP_ACCEPTED:
             return HTTPNotFound(request=req)
         return resp
