@@ -1,5 +1,5 @@
 # Copyright (c) 2010-2012 OpenStack Foundation
-# Copyright (c) 2016-2017 OpenIO SAS
+# Copyright (c) 2016-2018 OpenIO SAS
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,35 +14,45 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import json
 import mimetypes
 import time
 import math
 
 from swift import gettext_ as _
 from swift.common.utils import (
-    clean_content_type, config_true_value, Timestamp, public)
-from swift.common.constraints import check_metadata, check_object_creation, \
-    check_copy_from_header, check_destination_header, \
-    check_account_format
-from swift.common import constraints
-from swift.common.http import HTTP_CREATED, HTTP_MULTIPLE_CHOICES
+    clean_content_type, config_true_value, Timestamp, public,
+    close_if_possible, closing_if_possible)
+from swift.common.constraints import check_metadata, check_object_creation
+from swift.common.middleware.versioned_writes import DELETE_MARKER_CONTENT_TYPE
 from swift.common.swob import HTTPAccepted, HTTPBadRequest, HTTPNotFound, \
-    HTTPPreconditionFailed, HTTPRequestEntityTooLarge, HTTPRequestTimeout, \
-    HTTPUnprocessableEntity, HTTPClientDisconnect, Request, HTTPCreated, \
+    HTTPConflict, HTTPPreconditionFailed, HTTPRequestTimeout, \
+    HTTPUnprocessableEntity, HTTPClientDisconnect, HTTPCreated, \
     HTTPNoContent, Response, HTTPInternalServerError, multi_range_iterator
-from swift.common.request_helpers import is_sys_or_user_meta, is_sys_meta, \
-    remove_items, copy_header_subset
-from swift.proxy.controllers.base import _set_object_info_cache, \
+from swift.common.request_helpers import is_sys_or_user_meta
+from swift.proxy.controllers.base import set_object_info_cache, \
         delay_denial, cors_validation
-from swift.proxy.controllers.obj import check_content_type, copy_headers_into
+from swift.proxy.controllers.obj import check_content_type
 
 from swift.proxy.controllers.obj import BaseObjectController as \
         BaseObjectController
 
 from oio.common import exceptions
-from oio.common.http import ranges_from_http_header, quote
-from oio.common.green import ClientReadTimeout
-from oioswift.utils import IterO
+from oio.common.http import ranges_from_http_header
+
+try:
+    from oio.common.exceptions import SourceReadTimeout
+except ImportError:
+    # TODO(FVE): remove when dependency is oio>=4.2.0
+    class SourceReadTimeout(exceptions.OioTimeout):
+        pass
+# TODO(FVE): GreenSourceReadTimeout should never be raised from oio!
+# Remove when dependency is oio>=4.2.0
+from oio.common.green import SourceReadTimeout as GreenSourceReadTimeout
+from oioswift.utils import check_if_none_match, \
+    handle_not_allowed, handle_oio_timeout, handle_service_busy
+
+SLO = 'x-static-large-object'
 
 
 class ObjectControllerRouter(object):
@@ -79,6 +89,36 @@ class StreamRangeIterator(object):
         return self.stream
 
 
+class ExpectedSizeReader(object):
+    """Only accept as a valid EOF an exact number of bytes received."""
+
+    def __init__(self, source, expected):
+        self.source = source
+        self.expected = expected
+        self.consumed = 0
+
+    def read(self, *args, **kwargs):
+        rc = self.source.read(*args, **kwargs)
+        if len(rc) == 0:
+            if self.consumed != self.expected:
+                raise exceptions.SourceReadError("Truncated input")
+        else:
+            self.consumed = self.consumed + len(rc)
+        return rc
+
+    def readline(self, *args, **kwargs):
+        rc = self.source.readline(*args, **kwargs)
+        if len(rc) == 0:
+            if self.consumed != self.expected:
+                raise exceptions.SourceReadError("Truncated input")
+        else:
+            self.consumed = self.consumed + len(rc)
+        return rc
+
+    def close(self):
+        return close_if_possible(self.source)
+
+
 class ObjectController(BaseObjectController):
     allowed_headers = {'content-disposition', 'content-encoding',
                        'x-delete-at', 'x-object-manifest',
@@ -98,13 +138,14 @@ class ObjectController(BaseObjectController):
         """Handle GET requests."""
         return self.GETorHEAD(req)
 
+    @handle_oio_timeout
+    @handle_service_busy
+    @check_if_none_match
     def GETorHEAD(self, req):
         """Handle HTTP GET or HEAD requests."""
         container_info = self.container_info(
             self.account_name, self.container_name, req)
-
         req.acl = container_info['read_acl']
-
         policy_index = req.headers.get('X-Backend-Storage-Policy-Index',
                                        container_info['storage_policy'])
         req.headers['X-Backend-Storage-Policy-Index'] = policy_index
@@ -117,8 +158,8 @@ class ObjectController(BaseObjectController):
             resp = self.get_object_head_resp(req)
         else:
             resp = self.get_object_fetch_resp(req)
-        _set_object_info_cache(self.app, req.environ, self.account_name,
-                               self.container_name, self.object_name, resp)
+        set_object_info_cache(self.app, req.environ, self.account_name,
+                              self.container_name, self.object_name, resp)
         if ';' in resp.headers.get('content-type', ''):
             resp.content_type = clean_content_type(
                 resp.headers['content-type'])
@@ -127,10 +168,12 @@ class ObjectController(BaseObjectController):
 
     def get_object_head_resp(self, req):
         storage = self.app.storage
+        oio_headers = {'X-oio-req-id': self.trans_id}
         try:
-            metadata = storage.object_show(self.account_name,
-                                           self.container_name,
-                                           self.object_name)
+            metadata = storage.object_show(
+                self.account_name, self.container_name, self.object_name,
+                version=req.environ.get('oio.query', {}).get('version'),
+                headers=oio_headers)
         except (exceptions.NoSuchObject, exceptions.NoSuchContainer):
             return HTTPNotFound(request=req)
 
@@ -143,11 +186,12 @@ class ObjectController(BaseObjectController):
             ranges = ranges_from_http_header(req.headers.get('Range'))
         else:
             ranges = None
+        oio_headers = {'X-oio-req-id': self.trans_id}
         try:
-            metadata, stream = storage.object_fetch(self.account_name,
-                                                    self.container_name,
-                                                    self.object_name,
-                                                    ranges=ranges)
+            metadata, stream = storage.object_fetch(
+                self.account_name, self.container_name, self.object_name,
+                ranges=ranges, headers=oio_headers,
+                version=req.environ.get('oio.query', {}).get('version'))
         except (exceptions.NoSuchObject, exceptions.NoSuchContainer):
             return HTTPNotFound(request=req)
         resp = self.make_object_response(req, metadata, stream, ranges=ranges)
@@ -162,15 +206,19 @@ class ObjectController(BaseObjectController):
         resp = Response(request=req, conditional_response=True,
                         conditional_etag=conditional_etag)
 
-        resp.headers['Content-Type'] = metadata.get(
-            'mime_type', 'application/octet-stream')
+        if config_true_value(metadata['deleted']):
+            resp.headers['Content-Type'] = DELETE_MARKER_CONTENT_TYPE
+        else:
+            resp.headers['Content-Type'] = metadata.get(
+                'mime_type', 'application/octet-stream')
         properties = metadata.get('properties')
         if properties:
             for k, v in properties.iteritems():
                 if is_sys_or_user_meta('object', k) or \
                         k.lower() in self.allowed_headers:
-                            resp.headers[str(k)] = v
+                    resp.headers[str(k)] = v
         resp.headers['etag'] = metadata['hash'].lower()
+        resp.headers['x-object-sysmeta-version-id'] = metadata['version']
         ts = Timestamp(metadata['ctime'])
         resp.last_modified = math.ceil(float(ts))
         if stream:
@@ -201,41 +249,28 @@ class ObjectController(BaseObjectController):
     @public
     @cors_validation
     @delay_denial
+    @handle_not_allowed
+    @handle_oio_timeout
+    @handle_service_busy
+    @check_if_none_match
     def POST(self, req):
         """HTTP POST request handler."""
-        if self.app.object_post_as_copy:
-            req.method = 'PUT'
-            req.path_info = '/v1/%s/%s/%s' % (
-                self.account_name, self.container_name, self.object_name)
-            req.headers['Content-Length'] = 0
-            req.headers['X-Copy-From'] = quote('/%s/%s' % (self.container_name,
-                                                           self.object_name))
-            req.environ['swift.post_as_copy'] = True
-            req.environ['swift_versioned_copy'] = True
-            resp = self.PUT(req)
-            # Older editions returned 202 Accepted on object POSTs, so we'll
-            # convert any 201 Created responses to that for compatibility with
-            # picky clients.
-            if resp.status_int != HTTP_CREATED:
-                return resp
-            return HTTPAccepted(request=req)
-        else:
-            error_response = check_metadata(req, 'object')
-            if error_response:
-                return error_response
-            container_info = self.container_info(
-                self.account_name, self.container_name, req)
-            req.acl = container_info['write_acl']
-            if 'swift.authorize' in req.environ:
-                aresp = req.environ['swift.authorize'](req)
-                if aresp:
-                    return aresp
+        container_info = self.container_info(
+            self.account_name, self.container_name, req)
+        req.acl = container_info['write_acl']
+        if 'swift.authorize' in req.environ:
+            aresp = req.environ['swift.authorize'](req)
+            if aresp:
+                return aresp
+        error_response = check_metadata(req, 'object')
+        if error_response:
+            return error_response
 
-            policy_index = req.headers.get('X-Backend-Storage-Policy-Index',
-                                           container_info['storage_policy'])
-            stgpol = self._stgpol_from_policy_index(policy_index)
-            headers = self._prepare_headers(req)
-            return self._post_object(req, headers, stgpol)
+        policy_index = req.headers.get('X-Backend-Storage-Policy-Index',
+                                       container_info['storage_policy'])
+        stgpol = self._stgpol_from_policy_index(policy_index)
+        headers = self._prepare_headers(req)
+        return self._post_object(req, headers, stgpol)
 
     def _stgpol_from_policy_index(self, policy_index):
         # TODO actually convert policy_index to oio stgpol
@@ -244,13 +279,11 @@ class ObjectController(BaseObjectController):
     def _post_object(self, req, headers, stgpol):
         # TODO do something with stgpol
         metadata = self.load_object_metadata(headers)
-
-        storage = self.app.storage
-
+        oio_headers = {'X-oio-req-id': self.trans_id}
         try:
-            storage.object_set_properties(
+            self.app.storage.object_set_properties(
                 self.account_name, self.container_name, self.object_name,
-                metadata, clear=True)
+                metadata, clear=True, headers=oio_headers)
         except (exceptions.NoSuchObject, exceptions.NoSuchContainer):
             return HTTPNotFound(request=req)
         resp = HTTPAccepted(request=req)
@@ -259,19 +292,19 @@ class ObjectController(BaseObjectController):
     @public
     @cors_validation
     @delay_denial
+    @handle_not_allowed
+    @handle_oio_timeout
+    @handle_service_busy
+    @check_if_none_match
     def PUT(self, req):
         """HTTP PUT request handler."""
-        if req.if_none_match is not None and '*' not in req.if_none_match:
-            # Sending an etag with if-none-match isn't currently supported
-            return HTTPBadRequest(request=req, content_type='text/plain',
-                                  body='If-None-Match only supports *')
-
         container_info = self.container_info(
             self.account_name, self.container_name, req)
 
         req.acl = container_info['write_acl']
         req.environ['swift_sync_key'] = container_info['sync_key']
 
+        # is request authorized
         if 'swift.authorize' in req.environ:
             aresp = req.environ['swift.authorize'](req)
             if aresp:
@@ -279,51 +312,180 @@ class ObjectController(BaseObjectController):
 
         self._update_content_type(req)
 
+        # check constraints on object name and request headers
         error_response = check_object_creation(req, self.object_name) or \
             check_content_type(req)
         if error_response:
             return error_response
 
+        if req.headers.get('Oio-Copy-From'):
+            return self._link_object(req)
+
         self._update_x_timestamp(req)
 
-        source_header = req.headers.get('X-Copy-From')
-        if source_header:
-            error_response, req, data_source, update_response = \
-                self._handle_copy_request(req)
-            if error_response:
-                return error_response
-        else:
-            data_source = req.environ['wsgi.input']
-
-            def update_response(req, resp): return resp
+        data_source = req.environ['wsgi.input']
+        if req.content_length:
+            data_source = ExpectedSizeReader(data_source, req.content_length)
 
         headers = self._prepare_headers(req)
-        resp = self._store_object(req, data_source, headers)
-        return update_response(req, resp)
+        with closing_if_possible(data_source):
+            resp = self._store_object(req, data_source, headers)
+        return resp
 
     def _prepare_headers(self, req):
         req.headers['X-Timestamp'] = Timestamp(time.time()).internal
         headers = self.generate_request_headers(req, additional=req.headers)
         return headers
 
-    def _store_object(self, req, data_source, headers):
-        # TODO deal with stgpol
-        content_type = req.headers.get('content-type', 'octet/stream')
+    def _get_auto_policy_from_size(self, content_length):
+        # the default stgpol has an offset of 0 so should always be choose
+        policy = None
+        for (name, offset) in self.app.oio_stgpol:
+            if offset <= content_length:
+                policy = name
+
+        return policy
+
+    def _link_object(self, req):
+        _, container, obj = req.headers['Oio-Copy-From'].split('/', 2)
+
+        from_account = req.headers.get('X-Copy-From-Account',
+                                       self.account_name)
+        self.app.logger.info("LINK (%s,%s,%s) TO (%s,%s,%s)",
+                             from_account, self.container_name,
+                             self.object_name,
+                             self.account_name, container, obj)
         storage = self.app.storage
 
-        metadata = self.load_object_metadata(headers)
-        # TODO actually support if-none-match
+        if req.headers.get('Range'):
+            raise Exception("Fast Copy with Range is unsupported")
+
+            ranges = ranges_from_http_header(req.headers.get('Range'))
+            if len(ranges) != 1:
+                raise HTTPInternalServerError(
+                    request=req, body="mutiple ranges unsupported")
+            ranges = ranges[0]
+        else:
+            ranges = None
+
+        oio_headers = {'X-oio-req-id': self.trans_id}
+        # FIXME(FVE): use object_show, cache in req.environ
+        props = storage.object_get_properties(from_account, container, obj)
+        if props['properties'].get(SLO, None):
+            raise Exception("Fast Copy with SLO is unsupported")
+
+            if ranges is None:
+                raise HTTPInternalServerError(request=req,
+                                              body="LINK a MPU requires range")
+            self.app.logger.debug("LINK, original object is a SLO")
+
+            # retrieve manifest
+            _, data = storage.object_fetch(from_account, container, obj)
+            manifest = json.loads("".join(data))
+            offset = 0
+            # identify segment to copy
+            for entry in manifest:
+                if (ranges[0] == offset and
+                        ranges[1] + 1 == offset + entry['bytes']):
+                    _, container, obj = entry['name'].split('/', 2)
+                    checksum = entry['hash']
+                    self.app.logger.info(
+                        "LINK SLO (%s,%s,%s) TO (%s,%s,%s)",
+                        from_account, self.container_name,
+                        self.object_name,
+                        self.account_name, container, obj)
+                    break
+                offset += entry['bytes']
+            else:
+                raise HTTPInternalServerError(
+                    request=req, body="no segment matching range")
+        else:
+            checksum = props['hash']
+            if ranges:
+                raise HTTPInternalServerError(
+                    request=req, body="no range supported with single object")
+
         try:
-            chunks, size, checksum = storage.object_create(
-                self.account_name, self.container_name,
-                obj_name=self.object_name, file_or_path=data_source,
-                mime_type=content_type,
-                etag=req.headers.get('etag', '').strip('"'), metadata=metadata)
+            # TODO check return code (values ?)
+            storage.object_fastcopy(
+                from_account, container, obj,
+                self.account_name, self.container_name, self.object_name,
+                headers=oio_headers)
+        # TODO(FVE): this exception catching block has to be refactored
+        # TODO check which ones are ok or make non sense
+        except exceptions.Conflict:
+            raise HTTPConflict(request=req)
         except exceptions.PreconditionFailed:
             raise HTTPPreconditionFailed(request=req)
-        except ClientReadTimeout as err:
+        except exceptions.SourceReadError:
+            req.client_disconnect = True
             self.app.logger.warning(
-                _('ERROR Client read timeout (%ss)'), err.seconds)
+                _('Client disconnected without sending last chunk'))
+            self.app.logger.increment('client_disconnects')
+            raise HTTPClientDisconnect(request=req)
+        except exceptions.EtagMismatch:
+            return HTTPUnprocessableEntity(request=req)
+        except (exceptions.ServiceBusy, exceptions.OioTimeout):
+            raise
+        except (exceptions.NoSuchContainer, exceptions.NotFound):
+            raise HTTPNotFound(request=req)
+        except exceptions.ClientException as err:
+            # 481 = CODE_POLICY_NOT_SATISFIABLE
+            if err.status == 481:
+                raise exceptions.ServiceBusy()
+            self.app.logger.exception(
+                _('ERROR Exception transferring data %s'),
+                {'path': req.path})
+            raise HTTPInternalServerError(request=req)
+        except Exception:
+            self.app.logger.exception(
+                _('ERROR Exception transferring data %s'),
+                {'path': req.path})
+            raise HTTPInternalServerError(request=req)
+
+        resp = HTTPCreated(request=req, etag=checksum)
+        return resp
+
+    def _store_object(self, req, data_source, headers):
+        content_type = req.headers.get('content-type', 'octet/stream')
+        storage = self.app.storage
+        policy = None
+        container_info = self.container_info(self.account_name,
+                                             self.container_name, req)
+        if 'X-Oio-Storage-Policy' in req.headers:
+            policy = req.headers.get('X-Oio-Storage-Policy')
+            if not self.app.POLICIES.get_by_name(policy):
+                raise HTTPBadRequest(
+                    "invalid policy '%s', must be in %s" %
+                    (policy, self.app.POLICIES.by_name.keys()))
+        else:
+            try:
+                policy_index = int(
+                    req.headers.get('X-Backend-Storage-Policy-Index',
+                                    container_info['storage_policy']))
+            except TypeError:
+                policy_index = 0
+            if policy_index != 0:
+                policy = self.app.POLICIES.get_by_index(policy_index).name
+            else:
+                content_length = int(req.headers.get('content-length', 0))
+                policy = self._get_auto_policy_from_size(content_length)
+
+        metadata = self.load_object_metadata(headers)
+        oio_headers = {'X-oio-req-id': self.trans_id}
+        try:
+            _chunks, _size, checksum = storage.object_create(
+                self.account_name, self.container_name,
+                obj_name=self.object_name, file_or_path=data_source,
+                mime_type=content_type, policy=policy, headers=oio_headers,
+                etag=req.headers.get('etag', '').strip('"'), metadata=metadata)
+        except exceptions.Conflict:
+            raise HTTPConflict(request=req)
+        except exceptions.PreconditionFailed:
+            raise HTTPPreconditionFailed(request=req)
+        except (SourceReadTimeout, GreenSourceReadTimeout) as err:
+            self.app.logger.warning(
+                _('ERROR Client read timeout (%s)'), err)
             self.app.logger.increment('client_timeouts')
             raise HTTPRequestTimeout(request=req)
         except exceptions.SourceReadError:
@@ -334,133 +496,32 @@ class ObjectController(BaseObjectController):
             raise HTTPClientDisconnect(request=req)
         except exceptions.EtagMismatch:
             return HTTPUnprocessableEntity(request=req)
-        except exceptions.OioTimeout:
+        except (exceptions.ServiceBusy, exceptions.OioTimeout):
+            raise
+        # TODO(FVE): exceptions.NotFound is never raised from oio
+        # Remove when dependency is oio>=4.2.0
+        except (exceptions.NoSuchContainer, exceptions.NotFound):
+            raise HTTPNotFound(request=req)
+        except exceptions.ClientException as err:
+            # 481 = CODE_POLICY_NOT_SATISFIABLE
+            if err.status == 481:
+                raise exceptions.ServiceBusy()
             self.app.logger.exception(
-                _('ERROR Exception causing client disconnect'))
-            raise HTTPClientDisconnect(request=req)
+                _('ERROR Exception transferring data %s'),
+                {'path': req.path})
+            raise HTTPInternalServerError(request=req)
         except Exception:
             self.app.logger.exception(
                 _('ERROR Exception transferring data %s'),
                 {'path': req.path})
             raise HTTPInternalServerError(request=req)
 
-        resp = HTTPCreated(request=req, etag=checksum)
+        # TODO(FVE): use the timestamp of the object.
+        # Unfortunately the oio-sds API does not return the actual ctime.
+        resp = HTTPCreated(request=req, etag=checksum,
+                           last_modified=int(time.time()))
         return resp
 
-    def _handle_copy_request(self, req):
-        # TODO rely on oio copy instead?
-        if req.environ.get('swift.orig_req_method', req.method) != 'POST':
-            req.environ.setdefault('swift.log_info', []).append(
-                'x-copy-from:%s' % req.headers['X-Copy-From'])
-        ver, acct, _rest = req.split_path(2, 3, True)
-        src_account_name = req.headers.get('X-Copy-From-Account', None)
-        if src_account_name:
-            src_account_name = check_account_format(req, src_account_name)
-        else:
-            src_account_name = acct
-        src_container_name, src_obj_name = check_copy_from_header(req)
-        source_header = '/%s/%s/%s/%s' % (
-            ver, src_account_name, src_container_name, src_obj_name)
-        source_req = req.copy_get()
-
-        # make sure the source request uses it's container_info
-        source_req.headers.pop('X-Backend-Storage-Policy-Index', None)
-        source_req.path_info = source_header
-        source_req.headers['X-Newest'] = 'true'
-        if 'swift.post_as_copy' in req.environ:
-            # We're COPYing one object over itself because of a POST; rely on
-            # the PUT for write authorization, don't require read authorization
-            source_req.environ['swift.authorize'] = lambda req: None
-            source_req.environ['swift.authorize_override'] = True
-
-        orig_obj_name = self.object_name
-        orig_container_name = self.container_name
-        orig_account_name = self.account_name
-        sink_req = Request.blank(req.path_info,
-                                 environ=req.environ, headers=req.headers)
-
-        self.object_name = src_obj_name
-        self.container_name = src_container_name
-        self.account_name = src_account_name
-
-        source_resp = self.GET(source_req)
-
-        # This gives middlewares a way to change the source; for example,
-        # this lets you COPY a SLO manifest and have the new object be the
-        # concatenation of the segments (like what a GET request gives
-        # the client), not a copy of the manifest file.
-        hook = req.environ.get(
-            'swift.copy_hook',
-            (lambda source_req, source_resp, sink_req: source_resp))
-        source_resp = hook(source_req, source_resp, sink_req)
-
-        # reset names
-        self.object_name = orig_obj_name
-        self.container_name = orig_container_name
-        self.account_name = orig_account_name
-
-        if source_resp.status_int >= HTTP_MULTIPLE_CHOICES:
-            # this is a bit of ugly code, but I'm willing to live with it
-            # until copy request handling moves to middleware
-            return source_resp, None, None, None
-        if source_resp.content_length is None:
-            # This indicates a transfer-encoding: chunked source object,
-            # which currently only happens because there are more than
-            # CONTAINER_LISTING_LIMIT segments in a segmented object. In
-            # this case, we're going to refuse to do the server-side copy.
-            raise HTTPRequestEntityTooLarge(request=req)
-        if source_resp.content_length > constraints.MAX_FILE_SIZE:
-            raise HTTPRequestEntityTooLarge(request=req)
-
-        data_source = IterO(source_resp.app_iter)
-        sink_req.content_length = source_resp.content_length
-        sink_req.etag = source_resp.etag
-
-        # we no longer need the X-Copy-From header
-        del sink_req.headers['X-Copy-From']
-        if 'X-Copy-From-Account' in sink_req.headers:
-            del sink_req.headers['X-Copy-From-Account']
-        if not req.content_type_manually_set:
-            sink_req.headers['Content-Type'] = \
-                source_resp.headers['Content-Type']
-
-        fresh_meta_flag = config_true_value(
-            sink_req.headers.get('x-fresh-metadata', 'false'))
-
-        if fresh_meta_flag or 'swift.post_as_copy' in sink_req.environ:
-            # post-as-copy: ignore new sysmeta, copy existing sysmeta
-            def condition(key):
-                return is_sys_meta('object', key)
-            remove_items(sink_req.headers, condition)
-            copy_header_subset(source_resp, sink_req, condition)
-        else:
-            # copy/update existing sysmeta and user meta
-            copy_headers_into(source_resp, sink_req)
-            copy_headers_into(req, sink_req)
-
-        # copy over x-static-large-object for POSTs and manifest copies
-        if 'X-Static-Large-Object' in source_resp.headers and \
-                (req.params.get('multipart-manifest') == 'get' or
-                 'swift.post_as_copy' in req.environ):
-            sink_req.headers['X-Static-Large-Object'] = \
-                source_resp.headers['X-Static-Large-Object']
-
-        req = sink_req
-
-        def update_response(req, resp):
-            acct, path = source_resp.environ['PATH_INFO'].split('/', 3)[2:4]
-            resp.headers['X-Copied-From-Account'] = quote(acct)
-            resp.headers['X-Copied-From'] = quote(path)
-            if 'last-modified' in source_resp.headers:
-                resp.headers['X-Copied-From-Last-Modified'] = \
-                    source_resp.headers['last-modified']
-            copy_headers_into(req, resp)
-            return resp
-
-        # this is a bit of ugly code, but I'm willing to live with it
-        # until copy request handling moves to middleware
-        return None, req, data_source, update_response
-
     def _update_content_type(self, req):
         # Sometimes the 'content-type' header exists, but is set to None.
         req.content_type_manually_set = True
@@ -492,6 +553,9 @@ class ObjectController(BaseObjectController):
     @public
     @cors_validation
     @delay_denial
+    @handle_not_allowed
+    @handle_oio_timeout
+    @handle_service_busy
     def DELETE(self, req):
         """HTTP DELETE request handler."""
         container_info = self.container_info(
@@ -512,10 +576,12 @@ class ObjectController(BaseObjectController):
 
     def _delete_object(self, req):
         storage = self.app.storage
-
+        oio_headers = {'X-oio-req-id': self.trans_id}
         try:
-            storage.object_delete(self.account_name, self.container_name,
-                                  self.object_name)
+            storage.object_delete(
+                self.account_name, self.container_name, self.object_name,
+                version=req.environ.get('oio.query', {}).get('version'),
+                headers=oio_headers)
         except exceptions.NoSuchContainer:
             return HTTPNotFound(request=req)
         except exceptions.NoSuchObject:
@@ -523,32 +589,3 @@ class ObjectController(BaseObjectController):
             pass
         resp = HTTPNoContent(request=req)
         return resp
-
-    @public
-    @cors_validation
-    @delay_denial
-    def COPY(self, req):
-        """HTTP COPY request handler."""
-        if not req.headers.get('Destination'):
-            return HTTPPreconditionFailed(request=req,
-                                          body='Destination header required')
-        dest_account = self.account_name
-        if 'Destination-Account' in req.headers:
-            dest_account = req.headers.get('Destination-Account')
-            dest_account = check_account_format(req, dest_account)
-            req.headers['X-Copy-From-Account'] = self.account_name
-            self.account_name = dest_account
-            del req.headers['Destination-Account']
-        dest_container, dest_object = check_destination_header(req)
-        source = '/%s/%s' % (self.container_name, self.object_name)
-        self.container_name = dest_container
-        self.object_name = dest_object
-        # re-write the existing request as a PUT instead of creating a new one
-        # since this one is already attached to the posthooklogger
-        req.method = 'PUT'
-        req.path_info = '/v1/%s/%s/%s' % \
-                        (dest_account, dest_container, dest_object)
-        req.headers['Content-Length'] = 0
-        req.headers['X-Copy-From'] = quote(source)
-        del req.headers['Destination']
-        return self.PUT(req)
